'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var React = require('react');
var React__default = _interopDefault(React);

/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */

function __rest(s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
}

function unwrapExports (x) {
	return x && x.__esModule && Object.prototype.hasOwnProperty.call(x, 'default') ? x['default'] : x;
}

function createCommonjsModule(fn, module) {
	return module = { exports: {} }, fn(module, module.exports), module.exports;
}

var lib = createCommonjsModule(function (module, exports) {

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var React = _interopDefault(React__default);

var globalPrefix = "id";
var lastId = 0;
function nextId(localPrefix) {
  lastId++;
  return "".concat(localPrefix || globalPrefix).concat(lastId);
}
var resetId = function resetId() {
  lastId = 0;
};
var setPrefix = function setPrefix(newPrefix) {
  globalPrefix = newPrefix;
};

var getIds = function getIds(count, prefix) {
  var ids = [];

  for (var i = 0; i < count; i++) {
    ids.push(nextId(prefix));
  }

  return ids;
};

function usePrevious(value) {
  var ref = React.useRef();
  React.useEffect(function () {
    ref.current = value;
  });
  return ref.current;
}

function useId() {
  var count = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : 1;
  var prefix = arguments.length > 1 ? arguments[1] : undefined;
  var idsListRef = React.useRef([]);
  var prevCount = usePrevious(count);
  var prevPrefix = usePrevious(prefix);

  if (count !== prevCount || prevPrefix !== prefix) {
    idsListRef.current = getIds(count, prefix);
  }

  return idsListRef.current;
}

exports.default = nextId;
exports.resetId = resetId;
exports.setPrefix = setPrefix;
exports.useId = useId;
});

unwrapExports(lib);
var lib_1 = lib.resetId;
var lib_2 = lib.setPrefix;
var lib_3 = lib.useId;

/* eslint-disable @typescript-eslint/no-explicit-any */
const pendingPromise = Promise.race.bind(Promise, []);
class AnimationControl {
    constructor() {
        this._cancel = undefined;
        this._onFinishPromise = undefined;
        this._onFinishAction = undefined;
        this.cancel = () => {
            try {
                this._cancel && this._cancel();
            }
            catch (e) {
                console.log('Error canceling animation', e);
            }
        };
        this.createOnFinishPromise = (animationFinishPromise) => {
            console.log('** CREATING ON FINISH');
            let hasCanceled = false;
            this._onFinishPromise = new Promise((fulfill, _reject) => {
                this._cancel = () => {
                    fulfill(pendingPromise());
                    console.log('** FULLFILLING CANCEL');
                    hasCanceled = true;
                };
                try {
                    animationFinishPromise
                        .then(() => {
                        console.log('** FULLFILLING PROMISEE!!!');
                        !hasCanceled && fulfill();
                    })
                        .then(() => {
                        if (this._onFinishAction && !hasCanceled) {
                            this._onFinishAction();
                        }
                    });
                }
                catch (e) {
                    _reject(e);
                }
            });
            return this._onFinishPromise;
        };
        this.setOnFinishAction = (action) => {
            this._onFinishAction = action;
        };
    }
}

/* eslint-disable @typescript-eslint/no-explicit-any */
const setStateForNewAction = (setEState, triggerState, visible) => {
    console.log('Setting new state for new action!!!!');
    setEState(current => {
        // increment action count
        const actionCount = current.actionCount + 1;
        // either cancel the existing animation ('restarting') or prep for new one ('initializing')
        const currentState = current.currentState === 'running' ? 'restarting' : 'initalizing';
        // set the flag recording if an animation has taken place to false
        const hasRunForCycle = false;
        // copy over unmounted child states b/c they wont be in play anymore
        // otherwise set child states to undefined for this action
        const childStates = Object.keys(current.childStates).reduce((acc, childId) => {
            const currentChildState = current.childStates[childId];
            if (currentChildState === 'unmounted') {
                return Object.assign(Object.assign({}, acc), { [childId]: 'unmounted' });
            }
            else {
                return Object.assign(Object.assign({}, acc), { [childId]: undefined });
            }
        }, {});
        return {
            actionCount,
            currentState,
            triggerState,
            hasRunForCycle,
            childStates,
            visible
        };
    });
};
const setStateForFinishedAction = (setEState) => {
    setEState(current => (Object.assign(Object.assign({}, current), { currentState: current.currentState === 'running' ? 'finished' : current.currentState })));
};
const setHasRunForActionCount = (setEState) => {
    setEState(current => (Object.assign(Object.assign({}, current), { hasRunForCycle: true })));
};
const setCurrentStateToRunningForActionCount = (setEState) => {
    setEState(current => (Object.assign(Object.assign({}, current), { currentState: 'running' })));
};
const setCurrentStateToFinishedForActionCount = (setEState) => {
    setEState(current => (Object.assign(Object.assign({}, current), { currentState: 'finished' })));
};
const setCurrentStateToUnmountedForActionCount = (setEState) => {
    setEState(current => (Object.assign(Object.assign({}, current), { currentState: 'unmounted' })));
};
const setCurrentStateToInitializingForActionCount = (setEState) => {
    setEState(current => (Object.assign(Object.assign({}, current), { currentState: 'initalizing' })));
};
const setChildStateForActionCount = (setEState) => (id, state) => {
    setEState(current => (Object.assign(Object.assign({}, current), { childStates: Object.assign(Object.assign({}, current.childStates), { [id]: state }) })));
};

const childrenMatch = (childrenOfInterest, statesToMatch, allChildren) => {
    const childrenToLookAt = Object.keys(allChildren).filter(childId => childrenOfInterest.includes(childId));
    if (childrenToLookAt.length === 0) {
        return false;
    }
    return childrenToLookAt.reduce((acc, childId) => {
        return acc && statesToMatch.includes(allChildren[childId]);
    }, true);
};
const Animate = ({ name, visible: visibleProp, triggerState, predicateState, when, children, unMountOnHide: _unMountOnHide, id, enterAfterParentStart, enterAfterParentFinish, exitAfterChildStart, exitAfterChildFinish, animationBinding }) => {
    const [eState, setEState] = React.useState({
        actionCount: 0,
        currentState: 'initalizing',
        hasRunForCycle: false,
        triggerState: triggerState || undefined,
        visible: false,
        childStates: {}
    });
    const visible = animationBinding && animationBinding.parentVisible === false ? false : visibleProp;
    // const visible = visibleProp;
    const [uuid] = lib_3();
    const [ref, setRef] = React.useState();
    const refId = ref ? ref.id : undefined;
    const unMountOnHide = _unMountOnHide === undefined ? true : _unMountOnHide;
    console.log(name, ': ', 'START------------ref', refId, 'currentState:', eState.currentState, 'child state', eState.childStates, 'incoming trigger', triggerState, 'old trigger', eState.triggerState, 'incoming visibility', visible, 'old visibility', eState.visible);
    const createAnimationControl = () => {
        const ac = new AnimationControl();
        ac.setOnFinishAction(() => {
            setStateForFinishedAction(setEState);
        });
        return ac;
    };
    // TODO enable setting of animation control
    const [animationControl, setAnimationControl] = React.useState(createAnimationControl());
    React.useEffect(() => {
        if (animationBinding) {
            console.log(name, ': ', 'Notifying parent of state :))', eState.currentState);
            animationBinding.notifyParentOfState(id || uuid, eState.currentState);
        }
    }, [eState.currentState]);
    React.useEffect(() => {
        return () => {
            if (animationBinding) {
                console.log(name, ': ', 'Unmounting from unmount action');
                animationBinding.notifyParentOfState(id || uuid, 'unmounted');
            }
        };
    }, ['onExit']);
    React.useEffect(() => {
        setAnimationControl(c => {
            c.cancel();
            return createAnimationControl();
        });
        setStateForNewAction(setEState, triggerState, visible);
    }, [JSON.stringify(triggerState), visible]);
    const animate = (node) => {
        return (when || []).reduce((acc, predicateAnimation) => {
            const { hasRun, ctx } = acc;
            if (hasRun) {
                return acc;
            }
            let shouldRun;
            const predicate = predicateAnimation[0];
            if (Array.isArray(predicate)) {
                shouldRun = predicate.reduce((accc, predicate) => {
                    return accc && predicate(predicateState, { triggerState, visible });
                }, true);
            }
            else {
                shouldRun = predicate(predicateState, { triggerState, visible });
            }
            if (shouldRun) {
                const animation = predicateAnimation[1];
                const animationResult = animation(ctx);
                if (animationResult) {
                    return { hasRun: true, ctx, animationResult };
                }
                else {
                    return Object.assign(Object.assign({}, acc), { hasRun: true, ctx });
                }
            }
            return acc;
        }, { hasRun: false, ctx: { node }, animationResult: null });
    };
    React.useEffect(() => {
        console.log(name, ': ', 'Updated eState action count', eState.actionCount);
    }, [eState.actionCount]);
    React.useEffect(() => {
        console.log(name, ': ', 'Updated ref', refId);
        if (!refId) {
            console.log(name, ': ', 'Updated to no ref', refId);
        }
    }, [refId]);
    const parentState = (animationBinding && animationBinding.parentState) || 'initalizing';
    const parentVisible = (animationBinding && animationBinding.parentVisible) || true;
    React.useEffect(() => {
        console.log(name, ': ', 'Updated parentState', parentState);
    }, [parentState]);
    React.useEffect(() => {
        console.log(name, ': ', 'Updated parentVisible', parentVisible);
    }, [parentVisible]);
    React.useEffect(() => {
        console.log(name, ': ', 'Updated children states', eState.childStates);
    }, [JSON.stringify(eState.childStates)]);
    /**
     * Run animations whenever there is a state change
     */
    React.useEffect(() => {
        console.log(name, ': ', 'Child states:', eState.childStates);
        if (eState.currentState === 'restarting') {
            return;
        }
        if (ref == null) {
            console.log(name, ': ', 'WOULD PAUSE / STOP IF COULD');
            if (animationControl.cancel) {
                console.log(name, ': ', 'Canceling animation');
                animationControl.cancel();
            }
            if (!visible) {
                setCurrentStateToUnmountedForActionCount(setEState);
            }
            console.log(name, ': ', 'Skipping running of animation b/c ref missing');
            return;
        }
        if (eState.hasRunForCycle === true) {
            console.log(name, ': ', 'Has already run for cycle. Not running animation');
            return;
        }
        if (visible &&
            enterAfterParentStart &&
            parentState !== 'running' &&
            parentState !== 'finished') {
            console.log(name, ': ', 'Waiting for parent to start');
            return;
        }
        if (!visible &&
            exitAfterChildStart &&
            exitAfterChildStart.length > 0 &&
            childrenMatch(exitAfterChildStart, 
            // if matches the states that come before the finished state
            [undefined, 'restarting', 'initalizing'], eState.childStates)) {
            console.log(name, ': ', 'Waiting for child to start');
            return;
        }
        if (visible && enterAfterParentFinish && parentState !== 'finished') {
            console.log(name, ': ', 'Waiting for parent to finish');
            return;
        }
        if (!visible &&
            exitAfterChildFinish &&
            exitAfterChildFinish.length > 0 &&
            childrenMatch(exitAfterChildFinish, 
            // if matches the states that come before the finished state
            [undefined, 'restarting', 'initalizing', 'running'], eState.childStates)) {
            console.log(name, ': ', 'Waiting for child to finish');
            return;
        }
        console.log(name, ': ', 'Running animation');
        setHasRunForActionCount(setEState);
        const { ctx: animationCtx, hasRun, animationResult } = animate(ref);
        console.log(name, ': ', 'Animation running with', hasRun, animationCtx);
        if (hasRun) {
            setCurrentStateToRunningForActionCount(setEState);
        }
        if (animationResult && animationResult.finished) {
            animationControl.createOnFinishPromise(animationResult.finished);
        }
        else {
            console.log(name, ': ', 'No finish promises found. Setting state to finished');
            setCurrentStateToFinishedForActionCount(setEState);
        }
    }, [
        eState.currentState,
        eState.actionCount,
        refId,
        parentState,
        JSON.stringify(eState.childStates)
    ]);
    React.useEffect(() => {
        console.log(name, ': ', 'Updated currentState', eState.currentState);
    }, [eState.currentState]);
    React.useEffect(() => {
        if (eState.currentState === 'restarting') {
            console.log(name, 'Lifecycle ohhhh maybe');
            setCurrentStateToInitializingForActionCount(setEState);
        }
    }, [eState.currentState]);
    if (eState.currentState === 'restarting') {
        console.log(name, 'Lifecycle ohhhh yeah');
        // Return null to unmount children and allow new animation to be in correct dom position
        // incase an animation applied a transform or similar
        return null;
    }
    if (ref == null && unMountOnHide && visible === false) {
        return null;
    }
    const setRefOfAnimateable = (ref) => {
        // TODO: for some reason null is returned whenever this component rerenders.
        // Possibly due to the cloneElement behavior.
        // This prevents knowing about child unmount events, which isn't a big deal
        // if using the Animateable component.
        // However, if run in uncontrolled mode, this could be a problem.
        if (ref == null) {
            return;
        }
        setRef(ref);
    };
    const realChildren = children
        ? React__default.cloneElement(children, {
            ref: setRefOfAnimateable,
            id: uuid,
            animationBinding: {
                notifyParentOfState: setChildStateForActionCount(setEState),
                parentState: eState.currentState,
                parentVisible: eState.visible
            }
        })
        : null;
    if (visible &&
        enterAfterParentStart &&
        parentState !== 'running' &&
        parentState !== 'finished') {
        console.log(name, ': ', 'Waiting for parent to start before entering');
        return null;
    }
    if (visible && enterAfterParentFinish && parentState !== 'finished') {
        console.log(name, ': ', 'Waiting for parent to finish before entering');
        return null;
    }
    if (eState.currentState === 'unmounted') {
        console.log(name, ': ', 'Unmounted. Returning null');
        return null;
    }
    if (!visible &&
        unMountOnHide &&
        eState.currentState === 'finished' &&
        // TODO change to only do a shallow compare
        // Guard against running one new trigger states.
        //  Often, on !visible states we are also in the 'finished' state
        //  doing an unmount at the beginning can kill the child animations
        JSON.stringify({ triggerState, visible }) ===
            JSON.stringify({ triggerState: eState.triggerState, visible: eState.visible })) {
        console.log(name, ': ', 'Unmounting b/c finished', triggerState, eState);
        if (eState.currentState === 'finished') {
            setCurrentStateToUnmountedForActionCount(setEState);
        }
        return null;
    }
    else {
        console.log(name, ': ', 'Showing children');
        return realChildren;
    }
};

/* eslint-disable react/prop-types */
const Animateable = React__default.forwardRef(function animateable(props, ref) {
    if (!props.id) {
        throw new Error('Missing id');
    }
    if (!props.animationBinding) {
        throw new Error('No animation binding prop found. This usually means this component (the animateable component) is not directly mounted under an animation component');
    }
    return (React__default.createElement("div", { id: props.id, ref: ref, className: props.className }, props.children && props.animationBinding
        ? props.children(props.animationBinding)
        : props.children
            ? props.children
            : null));
});

/* eslint-disable react/prop-types */
const createRouterComponents = (routers // eslint-disable-line
) => {
    return Object.keys(routers).reduce((acc, routerName) => {
        const r = routers[routerName];
        // eslint-disable-next-line
        const component = ({ children, uncontrolled }) => {
            const [state, setState] = React.useState(r.state);
            React.useEffect(() => {
                if (r && r.subscribe) {
                    r.subscribe(all => setState(all.current));
                }
                return;
            }, ['startup']);
            if (uncontrolled) {
                return children ? children(r) : null;
            }
            return state.visible ? React__default.createElement(React__default.Fragment, null, children) : null;
        };
        // eslint-disable-next-line
        const Link = (_a) => {
            var { children, action, simulated } = _a, actionOptions = __rest(_a, ["children", "action", "simulated"]);
            const actionKeys = Object.keys(actionOptions);
            if (!simulated && actionKeys.length > 0) {
                throw new Error(`The 'simulated' prop must be used when the action option(s) '${actionKeys.join(' ')}' are used}.`);
            }
            /**
             * Subscribe to all state changes
             */
            // eslint-disable-next-line
            const [_, setRouterState] = React.useState();
            React.useEffect(() => {
                if (r.manager.serializedStateStore) {
                    r.manager.serializedStateStore.subscribeToStateChanges(all => setRouterState(all));
                }
                return;
            }, ['startup']);
            const link = r.link(action);
            return simulated ? (React__default.createElement("a", { href: undefined, title: link, onClick: () => r[action](actionOptions) }, children)) : (React__default.createElement("a", { href: link, title: link }, children));
        };
        // eslint-disable-next-line
        const ToggleLink = (_a) => {
            var { children, simulated } = _a, actionOptions = __rest(_a, ["children", "simulated"]);
            const actionKeys = Object.keys(actionOptions);
            if (!simulated && actionKeys.length > 0) {
                throw new Error(`The 'simulated' prop must be used when the action option(s) '${actionKeys.join(' ')}' are used}.`);
            }
            /**
             * Subscribe to all state changes
             */
            // eslint-disable-next-line
            const [_, setRouterState] = React.useState();
            React.useEffect(() => {
                if (r.manager.serializedStateStore) {
                    r.manager.serializedStateStore.subscribeToStateChanges(all => setRouterState(all));
                }
                return;
            }, ['startup']);
            const action = r.state.visible ? 'hide' : 'show';
            const link = r.link(action);
            return simulated ? (React__default.createElement("a", { href: undefined, title: link, onClick: () => r[action](actionOptions) }, children)) : (React__default.createElement("a", { href: link, title: link }, children));
        };
        // eslint-disable-next-line
        const Animate$1 = ({ when, children, unMountOnHide, id, enterAfterParentStart, enterAfterParentFinish, exitAfterChildStart, exitAfterChildFinish, animationBinding }) => {
            const [_triggerState, setTriggerState] = React.useState({
                visible: r.state.visible,
                data: r.state.data,
                actionCount: r.state.actionCount
            });
            React.useEffect(() => {
                if (r && r.subscribe) {
                    r.subscribe(all => {
                        const state = all.current;
                        setTriggerState({
                            visible: state.visible,
                            data: state.data,
                            actionCount: state.actionCount
                        });
                    });
                }
                return;
            }, ['startup']);
            return (React__default.createElement(Animate, { name: r.name, visible: _triggerState.visible, triggerState: _triggerState, predicateState: r, when: when, unMountOnHide: unMountOnHide, id: id, enterAfterParentStart: enterAfterParentStart, enterAfterParentFinish: enterAfterParentFinish, exitAfterChildStart: exitAfterChildStart, exitAfterChildFinish: exitAfterChildFinish, animationBinding: animationBinding }, children));
        };
        const updated = Object.assign(component, { Link, ToggleLink, Animate: Animate$1 });
        return Object.assign(Object.assign({}, acc), { [routerName]: updated });
    }, {});
};

exports.createRouterComponents = createRouterComponents;
