/// <reference types="react" />
export declare type AnimationState = 'restarting' | 'initalizing' | 'running' | 'finished' | 'unmounted';
export declare type NotifyParentOfState = (id: string, state: AnimationState) => void;
export declare type AnimationBinding = {
    notifyParentOfState: NotifyParentOfState;
    parentState: AnimationState;
    parentVisible: boolean;
};
/**
 * The state of the animation control component
 */
export declare type CurrentState<TriggerState> = {
    actionCount: number;
    currentState: AnimationState;
    hasRunForCycle: boolean;
    triggerState: TriggerState;
    childStates: {
        [childId: string]: AnimationState | undefined;
    };
    visible: boolean;
};
export declare type Predicate = <PS extends any, TS extends any>(predicateState: any, // TODO replace with PS once excessively deep type problem is fixed
{ triggerState, visible }: {
    triggerState: any;
    visible: boolean;
}) => boolean;
export declare type Predicates = Array<Predicate>;
export declare type AnimationCtx = {
    node: HTMLElement;
};
export interface IAnimationResult {
    finished: Promise<any>;
}
export declare type AnimationRun = {
    hasRun: boolean;
    ctx: AnimationCtx;
    animationResult: AnimationResult;
};
export declare type AnimationResult = IAnimationResult | null;
export declare type PredicateAnimation = (ctx: AnimationCtx) => AnimationResult;
export declare type When = Array<[Predicates | Predicate, PredicateAnimation]>;
export interface AnimateProps<PS, TS> {
    name: string;
    visible: boolean;
    triggerState?: TS;
    predicateState?: PS;
    when?: When;
    children?: React.ReactElement;
    unMountOnHide?: boolean;
    unMountOnShow?: boolean;
    id?: string;
    animationBinding?: AnimationBinding;
    enterAfterParentStart?: boolean;
    enterAfterParentFinish?: boolean;
    exitAfterChildStart?: string[];
    exitAfterChildFinish?: string[];
}
export interface AnimateableProps {
    id?: string;
    className?: string;
    animationBinding?: AnimationBinding;
    children?: <P, T extends string>(animationBinding: AnimationBinding | undefined) => any;
}
